import torch
from torch.nn import *
from transformers import AutoTokenizer, AutoModel


class BaseModel(Module):

    def __init__(self, args):
        super(BaseModel, self).__init__()
        self.device = args.device
        self.args = args
        self.lm = AutoModel.from_pretrained(args.lm)

        # if args.freeze_bert:
        #     for params in self.lm.parameters():
        #         params.requires_grad = False

        self.label_set = args.label_set
        hidden_size = self.lm.config.hidden_size
        self.heads = Sequential(
            Dropout(0.2),
            Tanh(),
            Linear(hidden_size, 18)
        )

    def forward(self, batch):

        B, U, L = batch['input_ids'].shape
        input_ids = batch['input_ids'].view(B * U, L).to(self.device)
        attention_mask = batch['attention_mask'].view(B * U, L).to(self.device)
        if 'token_type_ids' in batch:
            token_type_ids = batch['token_type_ids'].view(B * U, L).to(self.device)
        else:
            token_type_ids = None

        # (BU,L,D)
        res = self.lm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            return_dict=True
        )  # odict_keys(['last_hidden_state', 'pooler_output', 'hidden_states'])
        # print(res.keys())
        last_hidden_state = res['last_hidden_state']  # (BU,L,D)
        cls = last_hidden_state[:, 0, :]  # BU,D

        logits = self.heads(cls)  # BU,C
        logits = logits.view(-1, 2)
        preds = torch.argmax(logits, dim=-1)
        preds = preds.reshape(B, U, 9)
        return logits, preds


class LSTMModel(Module):

    def __init__(self, args):
        super(LSTMModel, self).__init__()
        self.device = args.device
        self.args = args
        self.label_set = args.label_set

        self.lm = AutoModel.from_pretrained(args.lm)
        if args.freeze_bert:
            for params in self.lm.parameters():
                params.requires_grad = False
        #
        hidden_size = self.lm.config.hidden_size
        self.lstm = LSTM(input_size=hidden_size,
                         hidden_size=hidden_size,
                         num_layers=1,
                         bidirectional=True,
                         batch_first=True)
        self.heads = Sequential(
            Dropout(0.2),
            Tanh(),
            Linear(2 * hidden_size, 18)
        )

    def forward(self, batch):
        B, U, L = batch['input_ids'].shape
        input_ids = batch['input_ids'].view(B * U, L).to(self.device)
        attention_mask = batch['attention_mask'].view(B * U, L).to(self.device)
        if 'token_type_ids' in batch:
            token_type_ids = batch['token_type_ids'].view(B * U, L).to(self.device)
        else:
            token_type_ids = None

        # (BU,L,D)
        res = self.lm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            return_dict=True
        )  # odict_keys(['last_hidden_state', 'pooler_output', 'hidden_states'])
        # print(res.keys())
        last_hidden_state = res['last_hidden_state']  # (BU,L,D)
        cls = last_hidden_state[:, 0, :]  # BU,D
        cls = cls.view(B, U, -1)

        output, (hn, cn) = self.lstm(cls)
        actual_U = self.args.utterance
        lstm_output = output[:, -actual_U:, :]
        logits = self.heads(output)  # B,U,C
        logits = logits.view(-1, 2)
        preds = torch.argmax(logits, dim=-1)
        preds = preds.reshape(B, actual_U, 9)
        return logits, preds


class CNNModel(Module):

    def __init__(self, args):
        super(CNNModel, self).__init__()
        self.device = args.device
        self.args = args
        self.label_set = args.label_set

        self.lm = AutoModel.from_pretrained(args.lm)
        if args.freeze_bert:
            for params in self.lm.parameters():
                params.requires_grad = False
        #
        hidden_size = self.lm.config.hidden_size
        kernel_size = args.kernel_size
        self.cnns = ModuleList([
            Conv1d(in_channels=hidden_size, out_channels=256, kernel_size=(k,))
            for k in range(1, self.args.kernel_size + 1)
        ])

        self.heads = Sequential(
            Dropout(0.2),
            Tanh(),

            Dropout(0.2),
            Tanh(),
            Linear(kernel_size * 256, 18)

        )

    def forward(self, batch):
        B, U, L = batch['input_ids'].shape
        input_ids = batch['input_ids'].view(B * U, L).to(self.device)
        attention_mask = batch['attention_mask'].view(B * U, L).to(self.device)
        if 'token_type_ids' in batch:
            token_type_ids = batch['token_type_ids'].view(B * U, L).to(self.device)
        else:
            token_type_ids = None

        # (BU,L,D)
        res = self.lm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            return_dict=True
        )  # odict_keys(['last_hidden_state', 'pooler_output', 'hidden_states'])
        # print(res.keys())
        last_hidden_state = res['last_hidden_state']  # (BU,L,D)
        cls = last_hidden_state[:, 0, :]  # BU,D
        cls = cls.view(B, U, -1)

        actual_U = self.args.utterance

        final_vectors = []

        cls = torch.transpose(cls, dim0=1, dim1=2)  # B,D,U

        for i, cnn in enumerate(self.cnns):
            output = cnn(cls)
            trim_output = output[:, :, -actual_U:]
            trim_output = torch.transpose(trim_output, dim0=1, dim1=2)
            final_vectors.append(trim_output)

        final_vector = torch.concat(final_vectors, dim=-1)
        logits = self.heads(final_vector)  # B,U,C
        logits = logits.view(-1, 2)
        preds = torch.argmax(logits, dim=-1)
        preds = preds.reshape(B, actual_U, 9)
        return logits, preds


class CNNLSTMModel(Module):

    def __init__(self, args):
        super(CNNLSTMModel, self).__init__()
        self.device = args.device
        self.args = args
        self.label_set = args.label_set

        self.lm = AutoModel.from_pretrained(args.lm)
        if args.freeze_bert:
            for params in self.lm.parameters():
                params.requires_grad = False
        #
        hidden_size = self.lm.config.hidden_size
        self.lstm = LSTM(input_size=hidden_size,
                         hidden_size=hidden_size,
                         num_layers=1,
                         bidirectional=True,
                         batch_first=True)

        kernel_size = args.kernel_size
        self.cnns = ModuleList([
            Conv1d(in_channels=hidden_size, out_channels=256, kernel_size=(k,))
            for k in range(1, self.args.kernel_size + 1)
        ])

        self.heads = Sequential(
            Dropout(0.2),
            Tanh(),

            Dropout(0.2),
            Tanh(),
            Linear(2 * hidden_size + kernel_size * 256, 18)

        )

    def forward(self, batch):
        B, U, L = batch['input_ids'].shape
        input_ids = batch['input_ids'].view(B * U, L).to(self.device)
        attention_mask = batch['attention_mask'].view(B * U, L).to(self.device)
        if 'token_type_ids' in batch:
            token_type_ids = batch['token_type_ids'].view(B * U, L).to(self.device)
        else:
            token_type_ids = None

        # (BU,L,D)
        res = self.lm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            return_dict=True
        )  # odict_keys(['last_hidden_state', 'pooler_output', 'hidden_states'])
        # print(res.keys())
        last_hidden_state = res['last_hidden_state']  # (BU,L,D)
        cls = last_hidden_state[:, 0, :]  # BU,D
        cls = cls.view(B, U, -1)

        output, (hn, cn) = self.lstm(cls)
        actual_U = self.args.utterance
        lstm_output = output[:, -actual_U:, :]

        final_vectors = [lstm_output]

        cls = torch.transpose(cls, dim0=1, dim1=2)  # B,D,U

        for i, cnn in enumerate(self.cnns):
            output = cnn(cls)
            trim_output = output[:, :, -actual_U:]
            trim_output = torch.transpose(trim_output, dim0=1, dim1=2)
            final_vectors.append(trim_output)

        final_vector = torch.concat(final_vectors, dim=-1)
        logits = self.heads(final_vector)  # B,U,C
        logits = logits.view(-1, 2)
        preds = torch.argmax(logits, dim=-1)
        preds = preds.reshape(B, actual_U, 9)
        return logits, preds


def init_weights(m):
    if type(m) == Linear:
        torch.nn.init.xavier_uniform_(m.weight)
        m.bias.data.fill_(0.01)


class GCNLayer(Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features):
        super(GCNLayer, self).__init__()
        self.linear = Linear(in_features, out_features)
        self.nonlinear = Tanh()

    def init_weight(self):
        self.apply(init_weights)

    def forward(self, text, adj):
        hidden = self.linear(text)
        denom = torch.sum(adj, dim=2, keepdim=True) + 1
        output = torch.matmul(adj, hidden) / denom
        return output


class GCNLSTMModel(Module):

    def __init__(self, args):
        super(GCNLSTMModel, self).__init__()
        self.device = args.device
        self.args = args
        self.label_set = args.label_set

        self.lm = AutoModel.from_pretrained(args.lm)
        if args.freeze_bert:
            for params in self.lm.parameters():
                params.requires_grad = False
        #
        hidden_size = self.lm.config.hidden_size
        self.lstm = LSTM(input_size=hidden_size,
                         hidden_size=hidden_size,
                         num_layers=1,
                         bidirectional=True,
                         batch_first=True)

        self.gcn1 = GCNLayer(2 * hidden_size, hidden_size)
        self.gcn2 = GCNLayer(hidden_size, hidden_size)
        self.heads = Sequential(
            Dropout(0.2),
            Tanh(),
            Linear(4 * hidden_size, 18)
        )

    def forward(self, batch):
        B, U, L = batch['input_ids'].shape
        input_ids = batch['input_ids'].view(B * U, L).to(self.device)
        attention_mask = batch['attention_mask'].view(B * U, L).to(self.device)
        adj = batch['speaker_graph'][:, :U, :U].to(self.device)
        if 'token_type_ids' in batch:
            token_type_ids = batch['token_type_ids'].view(B * U, L).to(self.device)
        else:
            token_type_ids = None

        # (BU,L,D)
        res = self.lm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            return_dict=True
        )  # odict_keys(['last_hidden_state', 'pooler_output', 'hidden_states'])
        # print(res.keys())
        last_hidden_state = res['last_hidden_state']  # (BU,L,D)
        cls = last_hidden_state[:, 0, :]  # BU,D
        cls = cls.view(B, U, -1)

        lstm_output, (hn, cn) = self.lstm(cls)
        actual_U = self.args.utterance

        gcn1 = self.gcn1(lstm_output, adj)
        gcn2 = self.gcn2(gcn1, adj)

        lstm_output = lstm_output[:, -actual_U:, :]
        gcn1 = gcn1[:, -actual_U:, :]
        gcn2 = gcn2[:, -actual_U:, :]

        outputs = torch.concat([lstm_output, gcn1, gcn2], dim=-1)

        logits = self.heads(outputs)  # B,U,C
        logits = logits.view(-1, 2)
        preds = torch.argmax(logits, dim=-1)
        preds = preds.reshape(B, actual_U, 9)
        return logits, preds


class GCNCNNModel(Module):

    def __init__(self, args):
        super(GCNCNNModel, self).__init__()
        self.device = args.device
        self.args = args
        self.label_set = args.label_set

        self.lm = AutoModel.from_pretrained(args.lm)
        if args.freeze_bert:
            for params in self.lm.parameters():
                params.requires_grad = False
        #
        hidden_size = self.lm.config.hidden_size
        self.gcn1 = GCNLayer(hidden_size, hidden_size)
        self.gcn2 = GCNLayer(hidden_size, hidden_size)
        kernel_size = args.kernel_size
        self.cnns = ModuleList([
            Conv1d(in_channels=hidden_size, out_channels=256, kernel_size=(k,))
            for k in range(1, self.args.kernel_size + 1)
        ])

        self.heads = Sequential(
            Dropout(0.2),
            Tanh(),
            Linear(2 * hidden_size + kernel_size * 256, 18)

        )

    def forward(self, batch):
        B, U, L = batch['input_ids'].shape
        input_ids = batch['input_ids'].view(B * U, L).to(self.device)
        attention_mask = batch['attention_mask'].view(B * U, L).to(self.device)
        adj = batch['speaker_graph'][:, :U, :U].to(self.device)
        if 'token_type_ids' in batch:
            token_type_ids = batch['token_type_ids'].view(B * U, L).to(self.device)
        else:
            token_type_ids = None

        # (BU,L,D)
        res = self.lm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            return_dict=True
        )  # odict_keys(['last_hidden_state', 'pooler_output', 'hidden_states'])
        # print(res.keys())
        last_hidden_state = res['last_hidden_state']  # (BU,L,D)
        cls = last_hidden_state[:, 0, :]  # BU,D
        cls = cls.view(B, U, -1)

        actual_U = self.args.utterance

        gcn1 = self.gcn1(cls, adj)
        gcn2 = self.gcn2(gcn1, adj)

        gcn1 = gcn1[:, -actual_U:, :]
        gcn2 = gcn2[:, -actual_U:, :]

        final_vectors = [gcn1, gcn2]

        cls = torch.transpose(cls, dim0=1, dim1=2)  # B,D,U

        for i, cnn in enumerate(self.cnns):
            output = cnn(cls)
            trim_output = output[:, :, -actual_U:]
            trim_output = torch.transpose(trim_output, dim0=1, dim1=2)
            final_vectors.append(trim_output)

        final_vector = torch.concat(final_vectors, dim=-1)
        logits = self.heads(final_vector)  # B,U,C
        logits = logits.view(-1, 2)
        preds = torch.argmax(logits, dim=-1)
        preds = preds.reshape(B, actual_U, 9)
        return logits, preds


class GCNCNNLSTMModel(Module):

    def __init__(self, args):
        super(GCNCNNLSTMModel, self).__init__()
        self.device = args.device
        self.args = args
        self.label_set = args.label_set

        self.lm = AutoModel.from_pretrained(args.lm)
        if args.freeze_bert:
            for params in self.lm.parameters():
                params.requires_grad = False
        #
        hidden_size = self.lm.config.hidden_size
        self.lstm = LSTM(input_size=hidden_size,
                         hidden_size=hidden_size,
                         num_layers=1,
                         bidirectional=True,
                         batch_first=True)
        self.gcn1 = GCNLayer(2 * hidden_size, hidden_size)
        self.gcn2 = GCNLayer(hidden_size, hidden_size)
        kernel_size = args.kernel_size
        self.cnns = ModuleList([
            Conv1d(in_channels=hidden_size, out_channels=256, kernel_size=(k,))
            for k in range(1, self.args.kernel_size + 1)
        ])

        self.heads = Sequential(
            Dropout(0.2),
            Tanh(),
            Linear(4 * hidden_size + kernel_size * 256, 18)

        )

    def forward(self, batch):
        B, U, L = batch['input_ids'].shape
        input_ids = batch['input_ids'].view(B * U, L).to(self.device)
        attention_mask = batch['attention_mask'].view(B * U, L).to(self.device)
        adj = batch['speaker_graph'][:, :U, :U].to(self.device)
        if 'token_type_ids' in batch:
            token_type_ids = batch['token_type_ids'].view(B * U, L).to(self.device)
        else:
            token_type_ids = None

        # (BU,L,D)
        res = self.lm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            return_dict=True
        )  # odict_keys(['last_hidden_state', 'pooler_output', 'hidden_states'])
        # print(res.keys())
        last_hidden_state = res['last_hidden_state']  # (BU,L,D)
        cls = last_hidden_state[:, 0, :]  # BU,D
        cls = cls.view(B, U, -1)

        lstm_output, (hn, cn) = self.lstm(cls)
        actual_U = self.args.utterance

        gcn1 = self.gcn1(lstm_output, adj)
        gcn2 = self.gcn2(gcn1, adj)

        lstm_output = lstm_output[:, -actual_U:, :]
        gcn1 = gcn1[:, -actual_U:, :]
        gcn2 = gcn2[:, -actual_U:, :]

        final_vectors = [lstm_output, gcn1, gcn2]

        cls = torch.transpose(cls, dim0=1, dim1=2)  # B,D,U

        for i, cnn in enumerate(self.cnns):
            output = cnn(cls)
            trim_output = output[:, :, -actual_U:]
            trim_output = torch.transpose(trim_output, dim0=1, dim1=2)
            final_vectors.append(trim_output)

        final_vector = torch.concat(final_vectors, dim=-1)
        logits = self.heads(final_vector)  # B,U,C
        logits = logits.view(-1, 2)
        preds = torch.argmax(logits, dim=-1)
        preds = preds.reshape(B, actual_U, 9)
        return logits, preds


class TransformerModel(Module):

    def __init__(self, args):
        super(TransformerModel, self).__init__()
        self.device = args.device
        self.args = args
        self.label_set = args.label_set

        self.lm = AutoModel.from_pretrained(args.lm)
        if args.freeze_bert:
            for params in self.lm.parameters():
                params.requires_grad = False
        #
        hidden_size = self.lm.config.hidden_size

        self.trans = TransformerEncoder(
            TransformerEncoderLayer(self.lm.config.hidden_size, 8),
            num_layers=1
        )

        self.heads = Sequential(
            Dropout(0.2),
            Tanh(),
            Linear(hidden_size, 18)
        )

    def forward(self, batch):
        B, U, L = batch['input_ids'].shape
        input_ids = batch['input_ids'].view(B * U, L).to(self.device)
        attention_mask = batch['attention_mask'].view(B * U, L).to(self.device)
        if 'token_type_ids' in batch:
            token_type_ids = batch['token_type_ids'].view(B * U, L).to(self.device)
        else:
            token_type_ids = None

        # (BU,L,D)
        res = self.lm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            return_dict=True
        )  # odict_keys(['last_hidden_state', 'pooler_output', 'hidden_states'])
        # print(res.keys())
        last_hidden_state = res['last_hidden_state']  # (BU,L,D)
        cls = last_hidden_state[:, 0, :]  # BU,D
        cls = cls.view(B, U, -1)
        # print('cls', tuple(cls.shape))
        output = self.trans(cls)

        actual_U = self.args.utterance
        output = output[:, -actual_U:, :]
        # print('Output', tuple(output.shape))

        logits = self.heads(output)  # B,U,C
        logits = logits.view(-1, 2)
        preds = torch.argmax(logits, dim=-1)
        preds = preds.reshape(B, actual_U, 9)
        return logits, preds


class GCNTransformerModel(Module):

    def __init__(self, args):
        super(GCNTransformerModel, self).__init__()
        self.device = args.device
        self.args = args
        self.label_set = args.label_set

        self.lm = AutoModel.from_pretrained(args.lm)
        if args.freeze_bert:
            for params in self.lm.parameters():
                params.requires_grad = False
        #
        hidden_size = self.lm.config.hidden_size

        self.gcn1 = GCNLayer(hidden_size, hidden_size)
        self.gcn2 = GCNLayer(hidden_size, hidden_size)

        self.trans = TransformerEncoder(
            TransformerEncoderLayer(self.lm.config.hidden_size, 8),
            num_layers=1
        )

        self.heads = Sequential(
            Dropout(0.2),
            Tanh(),
            Linear(hidden_size * 3, 18)
        )

    def forward(self, batch):
        B, U, L = batch['input_ids'].shape
        input_ids = batch['input_ids'].view(B * U, L).to(self.device)
        attention_mask = batch['attention_mask'].view(B * U, L).to(self.device)
        adj = batch['speaker_graph'][:, :U, :U].to(self.device)
        if 'token_type_ids' in batch:
            token_type_ids = batch['token_type_ids'].view(B * U, L).to(self.device)
        else:
            token_type_ids = None

        # (BU,L,D)
        res = self.lm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            return_dict=True
        )  # odict_keys(['last_hidden_state', 'pooler_output', 'hidden_states'])
        # print(res.keys())
        last_hidden_state = res['last_hidden_state']  # (BU,L,D)
        cls = last_hidden_state[:, 0, :]  # BU,D
        cls = cls.view(B, U, -1)
        # print('cls', tuple(cls.shape))
        trans_output = self.trans(cls)
        # print('trans_output', tuple(trans_output.shape))

        gcn1 = self.gcn1(trans_output, adj)
        gcn2 = self.gcn2(gcn1, adj)

        actual_U = self.args.utterance
        output = trans_output[:, -actual_U:, :]
        gcn1 = gcn1[:, -actual_U:, :]
        gcn2 = gcn2[:, -actual_U:, :]

        final_vectors = [output, gcn1, gcn2]

        # print('Output', tuple(output.shape))
        final_vectors = torch.concat(final_vectors, dim=-1)
        logits = self.heads(final_vectors)  # B,U,C
        logits = logits.view(-1, 2)
        preds = torch.argmax(logits, dim=-1)
        preds = preds.reshape(B, actual_U, 9)
        return logits, preds
