import random

import torch

from utils import *
from collections import defaultdict


class DocumentDataset(Dataset):

    def __init__(self, files, tokenizer, args):
        super(DocumentDataset, self).__init__()
        self.args = args
        self.tokenizer = tokenizer
        self.label_set = args.label_set
        data = [load_json(file, log=False) for file in files]
        self.n_context = args.context
        self.n_utterance = args.utterance
        self.bert_length = args.bert_length
        self.documents = []

        self.utterances = []

        for doc in data:
            self.utterances += [x for x in doc['annotation'] if len(x['text']) > 0]

    def __len__(self):
        return len(self.utterances) // self.n_utterance

    def __getitem__(self, i):
        C = self.n_context
        N = self.n_utterance
        utterances = self.utterances[i * N:(i + 1) * N]

        if i == 0:
            context = [
                {'text': ['[MASK]']}
                for _ in range(C)
            ]
        else:
            context = self.utterances[i * N - C:i * N]
        texts = []
        speakers = []

        # Add context
        for utt in context:
            text = '[CLS]' + ' [SEP] '.join(utt['text'])
            texts.append(text)
            if 'speaker' in utt:
                speakers.append(utt['speaker'])
            else:
                speakers.append('default')
        targets = []
        for utt in utterances:
            text = '[CLS]' + ' [SEP] '.join(utt['text'])
            texts.append(text)
            if 'speaker' in utt:
                speakers.append(utt['speaker'])
            else:
                speakers.append('default')
            target = [
                int(label in utt['utterance_label'])
                for label in self.label_set
            ]
            targets.append(target)
        res = self.tokenizer(texts, padding='max_length', return_tensors='pt')
        for k, v in res.items():
            res[k] = v[:, :self.bert_length]
        res['target'] = torch.LongTensor(targets)

        NC = N + C
        speaker_graph = torch.zeros((NC, NC))
        for i in range(N - 1):
            speaker_graph[i][i + 1] = 1.0
            speaker_graph[i + 1][i] = 1.0
        speaker_map = defaultdict(set)
        for i, speaker in enumerate(speakers):
            speaker_map[speaker].add(i)
        for utt_ids in speaker_map.values():
            for i1 in utt_ids:
                for i2 in utt_ids:
                    speaker_graph[i1][i2] = 1.0
                    speaker_graph[i2][i1] = 1.0
        res['speaker_graph'] = speaker_graph
        return res

    @staticmethod
    def pack(items):
        batch = {
            k: torch.stack([item[k] for item in items], dim=0)
            for k in items[0].keys()
        }
        return batch


TENSOR_TYPE = {
    'input_ids': torch.LongTensor,
    'attention_mask': torch.FloatTensor,
    'token_type_ids': torch.LongTensor,
    'speaker_graph': torch.FloatTensor,
    'target': torch.LongTensor,

}
