import datetime
import torch
from utils import *
from mydataset import *
from model import *

logging.getLogger("transformers").setLevel(logging.ERROR)
# Statistics
# Utterance 39 197
# Sentence/U 1 22
# Word/S 1 92

LABEL_SET = ['BCS', 'ELC', 'ELD', 'ELR', 'EVL', 'EXA', 'PVD', 'PVR', 'STC']
MODEL_DICT = {
    'base': BaseModel,
    'lstm': LSTMModel,
    'gcnlstm': GCNLSTMModel,
    'gcncnnlstm': GCNCNNLSTMModel,
    'gcncnn': GCNCNNModel,
    'cnnlstm': CNNLSTMModel,
    'cnn': CNNModel,
    'trans': TransformerModel,
    'gcntrans': GCNTransformerModel
}
OPTIMIZER_DICT = {
    'sgd': SGD,
    'adadelta': Adadelta,
    'adam': Adam,
    'adamw': AdamW
}

random.seed(1)
torch.manual_seed(1)
torch.cuda.manual_seed(1)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--lm', type=str, default='bert-base-cased', help='Language model name')
    parser.add_argument('--model', type=str, default='base', choices=MODEL_DICT.keys(), help='Model name')
    parser.add_argument('--utterance', type=int, default=8, help='Number of utterances per doc')
    parser.add_argument('--batch_size', type=int, default=4, help='Batch size')
    parser.add_argument('--bert_length', type=int, default=128, help='Number of word pieces')
    parser.add_argument('--context', type=int, default=4, help='Number of context utterances')
    parser.add_argument('--freeze_bert', default=False, action='store_true', help='Number of word pieces')
    parser.add_argument('--lr', type=float, default=0.00005, help='Learning rate')
    parser.add_argument('--epoch', type=int, default=5, help='Learning rate')
    parser.add_argument('--optimizer', type=str, default='adam', choices=['sgd', 'adam', 'adadelta', 'adamw'])
    parser.add_argument('--acc_iter', type=int, default=1, help='Number of accumulated gradients')
    parser.add_argument('--progress', default=False, action='store_true', help='Show progress bar')
    parser.add_argument('--kernel_size', type=int, default=4, choices=[1, 2, 3, 4], help='Max kernel_size')
    return parser


def train(model, dl, optimizer, ce, args):
    model.train()
    bar = tqdm.tqdm(dl, desc='Training') if args.progress else dl
    for ib, batch in enumerate(bar):
        # for k, v in batch.items():
        #     print(k, v.shape)
        logits, preds = model(batch)
        targets = batch['target'].view(-1).to(args.device)
        loss = ce(logits, targets) / args.acc_iter
        loss.backward()

        if ib > 0 and ib % args.acc_iter == 0:
            optimizer.step()
            optimizer.zero_grad()
            if args.progress:
                bar.set_description(f'Training: Loss={loss.item():.4f}')


def metrics(all_goldens, all_predictions):
    res = OrderedDict()
    ps, rs, fs = [], [], []
    for i, label in enumerate(LABEL_SET):
        goldens = all_goldens[:, i]
        predictions = all_predictions[:, i]
        p = precision_score(goldens, predictions, labels=[1]) * 100.0
        r = recall_score(goldens, predictions, labels=[1]) * 100.0
        f = f1_score(goldens, predictions, labels=[1]) * 100.0
        ps.append(p)
        rs.append(r)
        fs.append(f)
        res[label] = [p, r, f]
    res['AVG'] = (sum(ps) / 9.0, sum(rs) / 9.0, sum(fs) / 9.0,)
    return res


def evaluate(model, dl, msg, args):
    model.eval()
    all_goldens = []
    all_predictions = []
    bar = tqdm.tqdm(dl, desc=msg) if args.progress else dl
    for batch in bar:
        logits, preds = model(batch)  # preds: B,U,9

        all_goldens.append(batch['target'].view(-1, 9).cpu())
        all_predictions.append(preds.view(-1, 9).cpu())
    all_goldens = torch.concat(all_goldens, dim=0)
    all_predictions = torch.concat(all_predictions, dim=0)
    perfs = metrics(all_goldens, all_predictions)
    print(msg, now())
    for k, (p, r, f) in perfs.items():
        print(f'{k} {p:5.2f} {r:5.2f} {f:5.2f}')


def now():
    return datetime.now().strftime("%d/%m/%Y %H:%M:%S")


def main(args):
    for k, v in args.__dict__.items():
        print(f'{k}:\t{v}')

    splits = [
        glob.glob('../data/splits/split-0/*.json'),
        glob.glob('../data/splits/split-1/*.json'),
        glob.glob('../data/splits/split-2/*.json'),
        glob.glob('../data/splits/split-3/*.json'),
        glob.glob('../data/splits/split-4/*.json'),
    ]

    train_files = splits[0] + splits[1] + splits[2]
    dev_files = splits[3]
    test_files = splits[4]

    tokenizer = AutoTokenizer.from_pretrained(args.lm)

    train_dl = DataLoader(
        DocumentDataset(train_files, tokenizer, args),
        num_workers=2,
        batch_size=args.batch_size,
        collate_fn=DocumentDataset.pack,
        shuffle=True
    )
    dev_dl = DataLoader(
        DocumentDataset(dev_files, tokenizer, args),
        num_workers=2,
        batch_size=args.batch_size,
        collate_fn=DocumentDataset.pack,
        shuffle=False
    )
    test_dl = DataLoader(
        DocumentDataset(test_files, tokenizer, args),
        num_workers=2,
        batch_size=args.batch_size,
        collate_fn=DocumentDataset.pack,
        shuffle=False
    )

    model = MODEL_DICT[args.model](args)
    model.to(args.device)
    params = [x for x in model.parameters() if x.requires_grad]
    optimizer = OPTIMIZER_DICT[args.optimizer](lr=args.lr, params=params)
    ce = CrossEntropyLoss()

    for epoch in range(args.epoch):
        print(f'Epoch {epoch} starts at {now()}')
        train(model, train_dl, optimizer, ce, args)
        with torch.no_grad():
            evaluate(model, dev_dl, f'Evaluate Dev @ {epoch}', args)
            evaluate(model, test_dl, f'Evaluate Test @ {epoch}', args)


if __name__ == '__main__':
    parser = get_parser()
    args = parser.parse_args()
    args.device = 'cuda' if torch.cuda.is_available() else 'cpu'
    args.label_set = LABEL_SET
    main(args)
