import collections
import json
import glob
import os
import csv
import re
import numpy
import numpy as np
from nltk import word_tokenize
from os.path import *

from sklearn.svm import *
from sklearn.linear_model import *
from sklearn.metrics import *
from sklearn.tree import *
from sklearn.ensemble import *
from sklearn.feature_extraction.text import TfidfVectorizer
from qwk import quadratic_weighted_kappa_score

import random


# Read data
def read_json(path):
    with open(path) as f:
        data = json.load(f)
    return data


def read_score(path):
    with open(path) as f:
        r = csv.DictReader(f)
        data = dict()
        for row in r:
            name = row['PST_ID']
            tid = (name[0], re.search(r'[0-9]{4}', name).group(0))
            data[tid] = int(row['FinalScore'])
    return data


OVERALL_SCORES = read_score('data/transcript-scores.csv')
# OVERALL_SCORES = read_score('data/final_score.csv')
print(OVERALL_SCORES)
print('Total scores: ', len(OVERALL_SCORES))


def read_coding_data(split_dir):
    all_data = []
    files = glob.glob1(split_dir, '*.json')
    # files = [x for x in files if x[:7] in OVERALL_SCORES]
    files = sorted(files)
    for file in files:
        file = join(split_dir, file)
        data = read_json(file)
        all_data.append(data)
    return all_data


def extract_code_frequency_feature(transcript, teacher=False, student=False, data_flip_rate=0.0):
    annotation = transcript['annotation']
    counter = collections.Counter()
    for utterance in annotation:
        if teacher:
            if utterance['speaker_group'] == 'teacher':
                counter.update(utterance['utterance_label'])
        if student:
            if utterance['speaker_group'] == 'student':
                counter.update(utterance['utterance_label'])
    labels = ['BCS', 'ELC', 'ELD', 'ELR', 'EVL', 'EXA', 'PVD', 'PVR', 'STC']
    frequencies = [counter[l] for l in labels]
    total = sum(frequencies)
    if total == 0:
        props = [0.0 for _ in labels]
    else:
        props = [x / total for x in frequencies]
    return props


def extract_code_bigram_feature(transcript, teacher=False, student=False):
    labels = ['BCS', 'ELC', 'ELD', 'ELR', 'EVL', 'EXA', 'PVD', 'PVR', 'STC']
    bigrams = [(x, y) for x, y in zip(labels, labels)]

    utterance_labels = [x['utterance_label'] for x in transcript['annotation']]
    speakers = [x['speaker_group'] for x in transcript['annotation']]
    all_pairs = []
    for uid in range(len(utterance_labels) - 1):
        selected = False
        current_labels = utterance_labels[uid]
        current_speaker = speakers[uid]
        next_labels = utterance_labels[uid + 1]
        next_speaker = speakers[uid + 1]
        if teacher and next_speaker == 'teacher':
            selected = True
        if student and next_speaker == 'student':
            selected = True
        if selected:
            for l1 in current_labels:
                for l2 in next_labels:
                    all_pairs.append((l1, l2))
    counter = collections.Counter()
    counter.update(all_pairs)

    frequencies = []

    for x in bigrams:
        if x in counter:
            frequencies.append(counter[x])
        else:
            frequencies.append(0)
    total = sum(frequencies)
    if total == 0:
        props = [0.0 for x in bigrams]
    else:
        props = [x / float(total) for x in frequencies]
    return props


def extract_code_trigram_feature(transcript, teacher=False, student=False):
    labels = ['BCS', 'ELC', 'ELD', 'ELR', 'EVL', 'EXA', 'PVD', 'PVR', 'STC']
    bigrams = [(x, y, z) for x, y, z in zip(labels, labels, labels)]

    utterance_labels = [x['utterance_label'] for x in transcript['annotation']]
    speakers = [x['speaker_group'] for x in transcript['annotation']]
    all_pairs = []
    for uid in range(len(utterance_labels) - 2):
        selected = False
        current_labels = utterance_labels[uid]
        current_speaker = speakers[uid]
        next_labels = utterance_labels[uid + 1]
        next_speaker = speakers[uid + 1]
        next2_labels = utterance_labels[uid + 2]

        if teacher and next_speaker == 'teacher':
            selected = True
        if student and next_speaker == 'student':
            selected = True
        if selected:
            for l1 in current_labels:
                for l2 in next_labels:
                    for l3 in next2_labels:
                        all_pairs.append((l1, l2, l3))
    counter = collections.Counter()
    counter.update(all_pairs)

    frequencies = []

    for x in bigrams:
        if x in counter:
            frequencies.append(counter[x])
        else:
            frequencies.append(0)
    total = sum(frequencies)
    if total == 0:
        props = [0.0 for x in bigrams]
    else:
        props = [x / float(total) for x in frequencies]
    return props


def extract_word_feature(transcript):
    # Dictionary
    # with open("resources/vocab_full.json") as f:
    with open("resources/vocab.json") as f:
        words = f.read().split('\n')
        vocab = {
            w: i
            for i, w in enumerate(words)
        }
    # Loop
    vect = numpy.zeros((len(vocab)))
    for utterance in transcript["annotation"]:
        sentences = utterance["text"]
        for sent in sentences:
            tokens = word_tokenize(sent.lower())
            for token in tokens:
                if token in vocab:
                    vect[vocab[token]] = 1.0
    return vect


def extract_tfidf_feature(transcript, teacher=False, student=False):
    doc_sentences = []
    for utterance in transcript['annotation']:
        if utterance['speaker_group'] == 'teacher' and teacher:
            doc_sentences += utterance['text']
        if utterance['speaker_group'] == 'student' and student:
            doc_sentences += utterance['text']
    doc_text = ' '.join(doc_sentences)

    v = get_tfidf_vectorizer()

    vector = v.transform([doc_text])
    vector = vector.todense()
    # print(vector.shape)
    return vector.tolist()[0]


scores = read_score('data/transcript-scores.csv')


def load_data(split_dir):
    coding_data = read_coding_data(split_dir)
    bert_data = read_json('resources/bert-base-cased.json')

    print(list(bert_data.keys()))
    results = []
    keys = list(bert_data.keys())
    for k in keys:
        tid = k[:7]
        if tid not in bert_data:
            bert_data[tid] = bert_data[k]
    for i, x in enumerate(coding_data):
        name = x['title']
        tid = (name[0], re.search(r'[0-9]{4}', name).group(0))
        if tid not in scores:
            # print('Can not find: ', x['title'])
            continue

        score = scores[tid] - 1
        coding_data[i]['score'] = int(score)
        coding_data[i]['feature'] = []
        coding_data[i]['tfidf'] = extract_tfidf_feature(x, teacher=True, student=True)
        coding_data[i]['bert'] = bert_data[name+'.json']
        # coding_data[i]['feature'] += extract_code_frequency_feature(x, teacher=True, student=True)
        coding_data[i]['student_code'] = extract_code_frequency_feature(x, student=True)
        coding_data[i]['teacher_code'] = extract_code_frequency_feature(x, teacher=True)
        # coding_data[i]['feature'] += extract_code_bigram_feature(x, teacher=True, student=True)
        coding_data[i]['teacher_bigram'] = extract_code_bigram_feature(x, student=True)
        coding_data[i]['student_bigram'] = extract_code_bigram_feature(x, teacher=True)

        coding_data[i]['teacher_trigram'] = extract_code_trigram_feature(x, student=True)
        coding_data[i]['student_trigram'] = extract_code_trigram_feature(x, teacher=True)
        coding_data[i]['PST_ID'] = tid

        results.append(coding_data[i])
    print('Loaded', len(results), 'from ', split_dir)
    return results


def convert_to_label(preds):
    labels = []
    for x in preds:
        if x < 0.5:
            labels.append(0)
        elif x < 1.5:
            labels.append(1)
        else:
            labels.append(2)
    return labels


def aggregate_features(sample, feature_set):
    features = []
    for feature in feature_set:
        features += sample[feature]
    return features


def train_evaluate(train, test, feature_set):
    seed = 1
    random.seed(seed)
    np.random.seed(seed)
    train_x = [aggregate_features(x, feature_set) for x in train]
    train_y = [x['score'] for x in train]
    test_x = [aggregate_features(x, feature_set) for x in test]
    test_y = [x['score'] for x in test]

    # SVM, max=33
    # model = SVC(kernel='linear', C=100, random_state=512)
    model = SVR(kernel='linear', C=1)

    # model = RidgeClassifier(alpha=0.00001)
    # model = RidgeCV(alphas=1)

    # model = LinearRegression(normalize=True)
    # model = LogisticRegressionCV(max_iter=1000)

    # model = DecisionTreeClassifier()
    # model = DecisionTreeRegressor()

    # model = RandomForestClassifier(10)

    model.fit(train_x, train_y)

    train_preds = model.predict(train_x)
    regression_test_preds = model.predict(test_x)

    # train_preds = convert_to_label(train_preds)
    classification_test_preds = convert_to_label(regression_test_preds)

    # metrics(train_y, train_preds, )
    # metrics(test_y, test_preds)

    return regression_test_preds.tolist(), classification_test_preds


def metrics(targets, predictions, r_predictions=None, calculate_cm=False):
    # print(len(targets), len(predictions))
    acc = accuracy_score(targets, predictions)
    qwk = quadratic_weighted_kappa_score(targets, predictions)
    # p = precision_score(targets, predictions, average='micro', zero_division=0)
    # r = recall_score(targets, predictions, average='micro', zero_division=0)
    # f = f1_score(targets, predictions, average='micro', zero_division=0)

    r2 = r2_score(targets, predictions)
    if r_predictions:
        mse = mean_squared_error(targets, r_predictions)
    else:
        mse = mean_squared_error(targets, predictions)
    acc *= 100.
    # p *= 100.
    # r *= 100.
    # f *= 100.

    print('{:10.2f}\t{:10.4f}\t{:10.4f}\t{:10.4f}'.format(acc, r2, qwk, mse))

    # print(f'| Acc\t{acc:.2f}')
    # print(f'| P/R/F\t{p:.2f}\t{r:.2f}\t{f:.2f}')
    # print(f'| QWK\t{qwk:.2f}')
    # print(f'| MSE\t{mse:.4f}')

    if calculate_cm:
        cm = confusion_matrix(targets, predictions)
        for row in cm:
            print('{}\t{}\t{}'.format(*row))
    return [acc, r2, qwk, mse]


tfidf_vectorizer = None


def get_tfidf_vectorizer():
    global tfidf_vectorizer

    # with open('resources/vocab.json', 'r') as f:
    with open('resources/vocab_full.json', 'r') as f:
        vocab = json.load(f)
    if tfidf_vectorizer is None:

        documents = [read_json(x) for x in glob.glob('data/splits/*/*.json')]
        all_texts = []
        for doc in documents:
            doc_sentences = []
            for ann in doc['annotation']:
                doc_sentences += ann['text']
            doc_text = ' '.join(doc_sentences)
            all_texts.append(doc_text)

        # tfidf_vectorizer = TfidfVectorizer(stop_words='english', vocabulary=vocab)
        tfidf_vectorizer = TfidfVectorizer()
        tfidf_vectorizer.fit(all_texts)

    return tfidf_vectorizer


def main(base_dir):
    # Load data
    splits = [load_data(f'{base_dir}/split-{i}') for i in range(5)]

    # for x in splits:
    #     print('Split: ', len(x))

    # Setting up cross-validation
    train_0 = splits[1] + splits[2] + splits[3] + splits[4]
    train_1 = splits[0] + splits[2] + splits[3] + splits[4]
    train_2 = splits[0] + splits[1] + splits[3] + splits[4]
    train_3 = splits[0] + splits[1] + splits[2] + splits[4]
    train_4 = splits[0] + splits[1] + splits[2] + splits[3]

    test_0 = splits[0]
    test_1 = splits[1]
    test_2 = splits[2]
    test_3 = splits[3]
    test_4 = splits[4]

    target_0 = [x['score'] for x in test_0]
    target_1 = [x['score'] for x in test_1]
    target_2 = [x['score'] for x in test_2]
    target_3 = [x['score'] for x in test_3]
    target_4 = [x['score'] for x in test_4]
    targets = target_0 + target_1 + target_2 + target_3 + target_4

    # Run validation

    random.seed(2)
    np.random.seed(2)

    feature_settings = [
        ['bert'],
        ['bert', 'student_code'],
        ['bert', 'teacher_code'],
        ['bert', 'student_code', 'teacher_code'],
        ['bert', 'student_bigram'],
        ['bert', 'teacher_bigram'],
        ['bert', 'student_bigram', 'teacher_bigram'],
        ['bert', 'teacher_code', 'teacher_bigram'],
        # ['tfidf'],
        # ['tfidf', 'student_code'],
        # ['tfidf', 'teacher_code'],
        # ['tfidf', 'student_code', 'teacher_code'],
        # ['tfidf', 'student_bigram'],
        # ['tfidf', 'teacher_bigram'],
        # ['tfidf', 'student_trigram'],
        # ['tfidf', 'teacher_trigram'],
        # ['tfidf', 'student_bigram', 'teacher_bigram'],
        # ['tfidf', 'teacher_bigramher_code', 'teacher_bigram'],

        # [ 'student_code'],
        # [ 'teacher_code'],
        # [ 'student_code', 'teacher_code'],
        # [ 'student_bigram'],
        # [ 'teacher_bigram'],
        #
        # ['student_trigram'],
        # ['teacher_trigram'],

        # ['teacher_code', 'teacher_bigram', 'student_trigram'],
        # ['tfidf', 'teacher_code'],
        # ['tfidf', 'teacher_bigram'],
        # ['tfidf', 'student_trigram'],
        # ['tfidf', 'teacher_bigram', 'student_trigram'],
        # ['tfidf', 'teacher_code', 'student_trigram'],
        # ['tfidf', 'teacher_code', 'teacher_bigram'],
        # ['tfidf', 'teacher_code', 'teacher_bigram', 'student_trigram'],
        # [ 'student_bigram', 'teacher_bigram'],
        # [ 'teacher_code', 'teacher_bigram'],
    ]
    #
    # test_doc_ids = [x['title'] for x in test_0 + test_1 + test_2 + test_3 + test_4]

    print('Accc\t         R-square\t  QWK   \t   MSE')
    for feature_set in feature_settings:
        # print(feature_set)
        r_pred_0, c_pred_0 = train_evaluate(train_0, test_0, feature_set)
        r_pred_1, c_pred_1 = train_evaluate(train_1, test_1, feature_set)
        r_pred_2, c_pred_2 = train_evaluate(train_2, test_2, feature_set)
        r_pred_3, c_pred_3 = train_evaluate(train_3, test_3, feature_set)
        r_pred_4, c_pred_4 = train_evaluate(train_4, test_4, feature_set)

        r_preds = r_pred_0 + r_pred_1 + r_pred_2 + r_pred_3 + r_pred_4
        c_preds = c_pred_0 + c_pred_1 + c_pred_2 + c_pred_3 + c_pred_4
        # print('-' * 40)
        perf = metrics(targets, c_preds, r_preds, calculate_cm=False)

        # if perf[0] > 56:
        #
        #     for did, target, pred in zip(test_doc_ids, targets, c_preds):
        #         print(did, target + 1, pred + 1)


if __name__ == '__main__':
    main('data/splits/')

    # train_evaluate()
    # for rate in [0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09,
    #              0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]:
    #     main(f'data/splits-{rate}')
