import json
import numpy as np
from numpy.random import random
import os
import glob
from sklearn.metrics import precision_score, recall_score, f1_score, cohen_kappa_score

LABEL_SET = ['BCS', 'ELC', 'ELD', 'ELR', 'EVL', 'EXA', 'PVD', 'PVR', 'STC']


def read_json(path):
    with open(path) as f:
        data = json.load(f)
    return data


def save_json(data, path):
    dir_path = os.path.dirname(path)
    os.makedirs(dir_path, exist_ok=True)
    with open(path, 'w') as f:
        json.dump(data, f, indent=2)


def flip_data(original_labels, flip_rate):
    new_labels = []
    props = random(size=(9,))
    for p, label in zip(props, LABEL_SET):
        ori = label in original_labels
        if p < flip_rate:
            ori = not ori
        if ori:
            new_labels.append(label)
    return new_labels


def flip_and_save(path, flip_rate):
    data = read_json(path)
    annotations = data['annotation']
    for i, utterance in enumerate(annotations):
        labels = utterance['utterance_label']
        new_labels = flip_data(labels, flip_rate)
        annotations[i]['utterance_label'] = new_labels
    data['annotation'] = annotations

    new_path = path.replace('/splits/', f'/splits-{flip_rate}/')
    save_json(data, new_path)


def flip_all():
    np.random.seed(1111)
    files = glob.glob('data/splits/*/*.json')
    for file in files:
        for rate in [0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09,
                     0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]:
            flip_and_save(file, rate, random_seed)


def metrics(golden_doc_labels, compare_doc_labels):
    f1_scores = []
    kappa_scores = []
    for i, label in enumerate(LABEL_SET):
        goldens = [int(label in utterance_labels) for utterance_labels in golden_doc_labels]
        compares = [int(label in utterance_labels) for utterance_labels in compare_doc_labels]
        f = f1_score(goldens, compares, labels=[1]) * 100.0
        f1_scores.append(f)

        k = cohen_kappa_score(goldens, compares)
        kappa_scores.append(k)

    return sum(f1_scores) / 9.0, sum(kappa_scores) / 9.0


def evaluate(golden_dir, compared_dir):
    golden_files = sorted(glob.glob(golden_dir + '*/*.json'))
    compared_files = sorted(glob.glob(compared_dir + '*/*.json'))

    assert len(golden_files) == len(compared_files)

    all_golden_labels, all_compare_labels = [], []

    for g, c in zip(golden_files, compared_files):
        gname = os.path.basename(g)
        cname = os.path.basename(c)
        assert gname == cname

        golden_annotation = read_json(g)
        golden_labels = [x['utterance_label'] for x in golden_annotation['annotation']]
        compare_annotation = read_json(c)
        compare_labels = [x['utterance_label'] for x in compare_annotation['annotation']]

        all_golden_labels += golden_labels
        all_compare_labels += compare_labels

    macro_f1, macro_kappa = metrics(all_golden_labels, all_compare_labels)

    print(macro_f1, macro_kappa)


def evaluate_all():
    evaluate('data/splits/', 'data/splits-0.01/')
    evaluate('data/splits/', 'data/splits-0.02/')
    evaluate('data/splits/', 'data/splits-0.03/')
    evaluate('data/splits/', 'data/splits-0.04/')
    evaluate('data/splits/', 'data/splits-0.05/')
    evaluate('data/splits/', 'data/splits-0.06/')
    evaluate('data/splits/', 'data/splits-0.07/')
    evaluate('data/splits/', 'data/splits-0.08/')
    evaluate('data/splits/', 'data/splits-0.09/')
    evaluate('data/splits/', 'data/splits-0.1/')
    evaluate('data/splits/', 'data/splits-0.2/')
    evaluate('data/splits/', 'data/splits-0.3/')
    evaluate('data/splits/', 'data/splits-0.4/')
    evaluate('data/splits/', 'data/splits-0.5/')
    evaluate('data/splits/', 'data/splits-0.6/')
    evaluate('data/splits/', 'data/splits-0.7/')
    evaluate('data/splits/', 'data/splits-0.8/')
    evaluate('data/splits/', 'data/splits-0.9/')
    evaluate('data/splits/', 'data/splits-1.0/')


if __name__ == '__main__':
    # flip_all()

    evaluate_all()
